// This program is free software: you can use, modify and/or redistribute it
// under the terms of the simplified BSD License. You should have received a
// copy of this license along this program. If not, see
// <http://www.opensource.org/licenses/bsd-license.html>.
//
// Copyright (C) 2017 Zhaoyi Meng <mzhy@ucla.edu>
// All rights reserved.


#include<iostream>
#include<fstream>
#include<cmath>
#include<cstdlib>
#include "MBO_S.h"
#include <cblas.h>
#include <vector>
using namespace std;


MBO_S::MBO_S(){};
MBO_S::~MBO_S(){};
//f is a rows by cols matrix, index is a rows by 1 matrix,
//we find the largest element in each row of f and put its index in Index.
void MBO_S::get_index(double *Index, double *f,int rows,int cols) {
    for (int i = 0; i<rows; i++) {
        Index[i] = 0;
        double max = f[i*cols];
        for (int j = 0; j<cols; j++) {
            if(max<f[i*cols+j]) {
                max = f[i*cols+j];
                Index[i] = j;
            }
        }
    }
}

// calculate the purity score between two vectors
double MBO_S::puritymeas(double * index, double * preindex, int len,
                         int num_class){
    double score = 0;
    double avgscore = 0;
    for(int i = 0; i< num_class; i++){
        int pos = 0;
        std::vector<double> ind;
        for (int j = 0; j<len; j++){
            if(index[j] == i){
                ind.push_back(pos);
            }
            pos++;
        }
        int count = 0;
        int vec_length = ind.size();
        for(int k = 0; k<vec_length; k++){
            int temp = ind.at(k);
            if(preindex[temp]==i)
                count++;
        }
        score = score + count;
    }
    avgscore = score/len;
    return avgscore;
}

void MBO_S::MBO_algorithm(double * index, double * X, double * Lambda, double *
                          fidelity, int N, int n, int rows_fidelity,
                          int M,int s, double dt, double mu, double iterNum) {
    // some constants
    double alpha = 1.0;
    double beta = 0.0;
    double gamma = -1.0;
    int incOne = 1;
    double theta = -dt/s;
    int len_u = N*n;
    int len_a = M*n;
    // initialize u^0
    // For each row of u^0, we randomly select one component to be 1 and others
    // to be 0. For the rows corresponding to fidelity points, we assign the
    // component corresponding to the known labels to be 1 and others 0.
    double * u0 = new double [N*n];
    double * preindex = new double [N];
    
    for (int i = 0; i<N*n; i++) {
        u0[i] = 0;
    }
    for (int i = 0; i<N; i++) {
        int randint = rand()%n;
        index[i] = randint;
        int x = i*n+randint;
        u0[x] = 1;
    }
    cblas_dcopy(N,index,incOne,preindex,incOne);

    for (int i = 0; i<rows_fidelity; i++) {
        for(int j = 0; j<n; j++){
            int temp = fidelity[2*i]*n+j;
            u0[temp]=0;
        }
        int temp2 = fidelity[2*i]*n+fidelity[2*i+1];
        u0[temp2] = 1;
    }
    
    // allocating memory
    double * a = new double [M*n];
    double * Denom = new double[M];
    double * d = new double[M*n];
    double * u = new double[N*n];
    double * y0 = new double [N*n];
    double * F = new double[N*n];
    
    // in explicit formula, a_k^{n+1} = (1-dt*lambda_k)a_k^n-dt*d_k^n,
    // Denom[i] = 1-dt*lambda_k
    for (int i =0; i<M; i++) {
        Denom[i] = 1.0-Lambda[i]*dt/s;
    }
    
    // initialize d^0, mu*lambda(x)*(u-\hat{u}) = d*X
    for (int i= 0; i<M*n; i++) {
        d[i] = 0;
    }
    // u = u0
    cblas_dcopy(len_u,u0,incOne,u,incOne);
    
    for (int iter=0; iter<iterNum;iter++) {
        cout << "iter"<<iter<<endl;
        // a = X^T*u
        cblas_dgemm(CblasRowMajor, CblasTrans, CblasNoTrans,
                    M, n, N, alpha, X, M, u, n, beta, a, n);
        // step 1: diffuse
        // we do step1 (diffuse) s times before step 2 (thresholding)
        for (int j = 0; j<s;j++) {
            // a = a*Denom-(dt/s)*d
            for (int i = 0; i<M; i++) {
                for (int j = 0; j<n;j++) {
                    a[i*n+j]=a[i*n+j]*Denom[i];
                }
            }
            cblas_daxpy(len_a,theta,d,incOne,a,incOne );
            
            // computing u^{n+1/2} = X*a;
            cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
                        N, n, M, alpha, X, M, a, n, beta, u, n);
            
            // computing d^n via (11)
            cblas_dcopy(len_u,u,incOne,y0,incOne);
            
            //u = u-u0;
            cblas_daxpy(len_u,gamma,u0,incOne,y0,incOne );
            
            // F = lambda(x)(u-u0}), lambda(x) = 1 if this pixel has
            // known label, and 0 otherwise
            for (int i = 0; i<N*n; i++) {
                F[i]=0;
            }
            for (int i = 0; i<rows_fidelity; i++) {
                for (int j = 0; j<n; j++) {
                    int temp = fidelity[i*2]*n+j;
                    F[temp] = y0[temp];
                }
            }
            // d is the coefficient of mu*lambda(x)*(u-u0) projected onto X,
            // mu*F = X*d, so d = mu*X^T*F
            cblas_dgemm(CblasRowMajor, CblasTrans, CblasNoTrans,
                        M, n, N, alpha, X, M, F, n, beta, d, n);
            for (int i = 0; i<M*n; i++) {
                d[i] = d[i]*mu;
            }
        }
        
        //step 2: thresholding part
        // u_i^{n+1} = e_r, where e_r = argmax(u_i^{n+1/2})

        get_index(index, u,N,n);
        
        for (int i = 0; i<N*n; i++) {
            u[i]=0;
        }
        
        for (int i = 0; i<N; i++) {
            int x = i*n+index[i];
            u[x]= 1;
        }
        
        // criteria to stop
        double purity_score = puritymeas(index,preindex, N, n);
        if (purity_score>0.9999) {
            break;
        }
        
        // preindex = index
        cblas_dcopy(N,index,incOne,preindex,incOne);
    }
    
    delete [] a;
    delete [] Denom;
    delete [] d;
    delete [] u;
    delete [] y0;
    delete [] F;
    delete [] u0;
    
}
